@@ -80,6 +80,9 @@ CIFSSMBNegotiate(unsigned int xid, struct cifsSesInfo *ses,
 	pSMB->hdr.Flags2 |= SMBFLG2_UNICODE;
 	if (extended_security)
 		pSMB->hdr.Flags2 |= SMBFLG2_EXT_SEC;
+    if (sign_CIFS_PDUs) {
+        pSMB->hdr.Flags2 |= SMBFLG2_SECURITY_SIGNATURE;
+    }
 
 	pSMB->ByteCount = strlen(protocols[0].name) + 1;
 	strncpy(pSMB->DialectsArray, protocols[0].name, 30);	/* null guaranteed to be at end of source and target buffers anyway */
@@ -145,6 +148,11 @@ CIFSSMBNegotiate(unsigned int xid, struct cifsSesInfo *ses,
 
 		} else
 			ses->capabilities &= ~CAP_EXTENDED_SECURITY;
+        if(sign_CIFS_PDUs == FALSE) {        
+            if(ses->secMode & SECMODE_SIGN_REQUIRED)
+                cERROR(1,("\nServer required CIFS packet signing - enable /proc/fs/cifs/PacketSigningEnabled"));
+            ses->secMode &= ~(SECMODE_SIGN_ENABLED | SECMODE_SIGN_REQUIRED);
+        }
 	}
 	if (pSMB)
 		buf_release(pSMB);
@@ -222,7 +230,8 @@ CIFSSMBLogoff(const int xid, struct cifsSesInfo *ses)
 		up(&ses->sesSem);
 		return -EBUSY;
 	}
-
+    if(ses->secMode & (SECMODE_SIGN_REQUIRED | SECMODE_SIGN_ENABLED))
+        pSMB->hdr.Flags2 |= SMBFLG2_SECURITY_SIGNATURE;
 	rc = smb_init(SMB_COM_LOGOFF_ANDX, 2, 0 /* no tcon anymore */ ,
 		      (void **) &pSMB, (void **) &smb_buffer_response);
 	if (rc) {
@@ -233,7 +242,6 @@ CIFSSMBLogoff(const int xid, struct cifsSesInfo *ses)
 	pSMB->hdr.Uid = ses->Suid;
 
 	pSMB->AndXCommand = 0xFF;
-
 	rc = SendReceive(xid, ses, (struct smb_hdr *) pSMB,
 			 smb_buffer_response, &length, 0);
 	if (ses->server) {
@@ -388,8 +396,6 @@ CIFSSMBMkDir(const int xid, const struct cifsTconInfo *tcon,
 	return rc;
 }
 
-/* BB Need to continue endian neutral code fixing from here down and in SessSetup and tConx */
-
 int
 CIFSSMBOpen(const int xid, const struct cifsTconInfo *tcon,
 	    const char *fileName, const int openDisposition,
@@ -432,8 +438,10 @@ CIFSSMBOpen(const int xid, const struct cifsTconInfo *tcon,
 	pSMB->DesiredAccess = cpu_to_le32(access_flags);
 	pSMB->AllocationSize = 0;
 	pSMB->FileAttributes = ATTR_NORMAL;	/* XP does not handle ATTR_POSIX_SEMANTICS */
-	if ((omode & S_IWUGO) == 0)
-		pSMB->FileAttributes |= ATTR_READONLY;	/* BB might as well do this until CIFS ACL support is put in */
+	/*if ((omode & S_IWUGO) == 0)
+		pSMB->FileAttributes |= ATTR_READONLY;*/	
+    /*  Above line causes problems due to problem with vfs splitting create into
+        two pieces - need to set mode after file created not while it is being created */
 	pSMB->FileAttributes = cpu_to_le32(pSMB->FileAttributes);
 	pSMB->ShareAccess = cpu_to_le32(FILE_SHARE_ALL);
 	pSMB->CreateDisposition = cpu_to_le32(openDisposition);
@@ -1536,6 +1544,9 @@ CIFSGetDFSRefer(const int xid, struct cifsSesInfo *ses,
 	if (ses->capabilities & CAP_DFS) {
 		pSMB->hdr.Flags2 |= SMBFLG2_DFS;
 	}
+    if(ses->secMode & (SECMODE_SIGN_REQUIRED | SECMODE_SIGN_ENABLED))
+       pSMB->hdr.Flags2 |= SMBFLG2_SECURITY_SIGNATURE;
+
 	if (ses->capabilities & CAP_UNICODE) {
 		pSMB->hdr.Flags2 |= SMBFLG2_UNICODE;
 		name_len =
@@ -1863,14 +1874,19 @@ CIFSSMBQFSUnixInfo(int xid, struct cifsTconInfo *tcon,
 	return rc;
 }
 
+/* We can not use write of zero bytes trick to 
+   set file size due to need for large file support.  Also note that 
+   this SetPathInfo is preferred to SetFileInfo based method in next 
+   routine which is only needed to work around a sharing violation bug
+   in Samba which this routine can run into */
+
 int
 CIFSSMBSetEOF(int xid, struct cifsTconInfo *tcon, char *fileName,
-	      __u64 size, const struct nls_table *nls_codepage)
+	      __u64 size, int SetAllocation, const struct nls_table *nls_codepage)
 {
-/* level 0x104 T2SetPathInfo - we can not use write of zero bytes trick to set file size due to need for large file support*/
-	TRANSACTION2_SPI_REQ *pSMB = NULL;
-	TRANSACTION2_SPI_RSP *pSMBr = NULL;
-	FILE_END_OF_FILE_INFO *parm_data;
+	struct smb_com_transaction2_spi_req *pSMB = NULL;
+	struct smb_com_transaction2_spi_rsp *pSMBr = NULL;
+	struct file_end_of_file_info *parm_data;
 	int name_len;
 	int rc = 0;
 	int bytes_returned = 0;
@@ -1896,7 +1912,7 @@ CIFSSMBSetEOF(int xid, struct cifsTconInfo *tcon, char *fileName,
 		strncpy(pSMB->FileName, fileName, name_len);
 	}
 	pSMB->ParameterCount = 6 + name_len;
-	pSMB->DataCount = sizeof (FILE_END_OF_FILE_INFO);
+	pSMB->DataCount = sizeof (struct file_end_of_file_info);
 	pSMB->MaxParameterCount = cpu_to_le16(2);
 	pSMB->MaxDataCount = cpu_to_le16(1000);	/* BB find exact max SMB PDU from sess structure BB */
 	pSMB->MaxSetupCount = 0;
@@ -1907,8 +1923,24 @@ CIFSSMBSetEOF(int xid, struct cifsTconInfo *tcon, char *fileName,
 	pSMB->ParameterOffset =
 	    (int) &(pSMB->InformationLevel) - (int) pSMB->hdr.Protocol;
 	pSMB->DataOffset = pSMB->ParameterOffset + pSMB->ParameterCount;
+    if(SetAllocation) {
+        if (tcon->ses->capabilities & CAP_INFOLEVEL_PASSTHRU)
+            pSMB->InformationLevel =
+                cpu_to_le16(SMB_SET_FILE_ALLOCATION_INFO2);
+        else
+            pSMB->InformationLevel =
+                cpu_to_le16(SMB_SET_FILE_ALLOCATION_INFO);
+    } else /* Set File Size */  {    
+	    if (tcon->ses->capabilities & CAP_INFOLEVEL_PASSTHRU)
+		    pSMB->InformationLevel =
+		        cpu_to_le16(SMB_SET_FILE_END_OF_FILE_INFO2);
+	    else
+		    pSMB->InformationLevel =
+		        cpu_to_le16(SMB_SET_FILE_END_OF_FILE_INFO);
+    }
+
 	parm_data =
-	    (FILE_END_OF_FILE_INFO *) (((char *) &pSMB->hdr.Protocol) +
+	    (struct file_end_of_file_info *) (((char *) &pSMB->hdr.Protocol) +
 				       pSMB->DataOffset);
 	pSMB->ParameterOffset = cpu_to_le16(pSMB->ParameterOffset);
 	pSMB->DataOffset = cpu_to_le16(pSMB->DataOffset);
@@ -1920,12 +1952,6 @@ CIFSSMBSetEOF(int xid, struct cifsTconInfo *tcon, char *fileName,
 	pSMB->TotalDataCount = pSMB->DataCount;
 	pSMB->ParameterCount = cpu_to_le16(pSMB->ParameterCount);
 	pSMB->TotalParameterCount = pSMB->ParameterCount;
-	if (tcon->ses->capabilities & CAP_INFOLEVEL_PASSTHRU)
-		pSMB->InformationLevel =
-		    cpu_to_le16(SMB_SET_FILE_END_OF_FILE_INFO2);
-	else
-		pSMB->InformationLevel =
-		    cpu_to_le16(SMB_SET_FILE_END_OF_FILE_INFO);
 	pSMB->Reserved4 = 0;
 	pSMB->hdr.smb_buf_length += pSMB->ByteCount;
 	parm_data->FileSize = cpu_to_le64(size);
@@ -1944,76 +1970,82 @@ CIFSSMBSetEOF(int xid, struct cifsTconInfo *tcon, char *fileName,
 }
 
 int
-CIFSSMBSetAllocationSize(int xid, struct cifsTconInfo *tcon,
-			 char *fileName, __u64 size,
-			 const struct nls_table *nls_codepage)
+CIFSSMBSetFileSize(const int xid, struct cifsTconInfo *tcon, __u64 size, 
+                   __u16 fid, __u32 pid_of_opener, int SetAllocation)
 {
-/* level 0x104 T2SetPathInfo - we can not use write of zero bytes trick to set file size due to need for large file support*/
-	TRANSACTION2_SPI_REQ *pSMB = NULL;
-	TRANSACTION2_SPI_RSP *pSMBr = NULL;
-	FILE_END_OF_FILE_INFO *parm_data;
-	int name_len;
+	struct smb_com_transaction2_sfi_req *pSMB  = NULL;
+	struct smb_com_transaction2_sfi_rsp *pSMBr = NULL;
+	char *data_offset;
+    struct file_end_of_file_info *parm_data;
 	int rc = 0;
 	int bytes_returned = 0;
+    __u32 tmp;
+
+	cFYI(1, ("\nSetFileSize (via SetFileInfo)"));
 
 	rc = smb_init(SMB_COM_TRANSACTION2, 15, tcon, (void **) &pSMB,
 		      (void **) &pSMBr);
 	if (rc)
 		return rc;
-
-	if (pSMB->hdr.Flags2 & SMBFLG2_UNICODE) {
-		name_len =
-		    cifs_strtoUCS((wchar_t *) pSMB->FileName, fileName, 530
-				  /* find define for this maxpathcomponent */
-				  , nls_codepage);
-		name_len++;	/* trailing null */
-		name_len *= 2;
-	} else {		/* BB improve the check for buffer overruns BB */
-
-		name_len = strnlen(fileName, 530);
-		name_len++;	/* trailing null */
-		strncpy(pSMB->FileName, fileName, name_len);
-	}
-	pSMB->ParameterCount = 6 + name_len;
-	pSMB->DataCount = sizeof (FILE_END_OF_FILE_INFO);
-	pSMB->MaxParameterCount = cpu_to_le16(2);
-	pSMB->MaxDataCount = cpu_to_le16(1000);	/* BB find exact max SMB PDU from sess structure BB */
+	tmp = cpu_to_le32(current->pid);  /* override pid of current process
+                                         so network fid will be valid */
+	pSMB->hdr.Pid = tmp & 0xFFFF;
+	tmp >>= 16;
+	pSMB->hdr.PidHigh = tmp & 0xFFFF;
+    
+	pSMB->ParameterCount = 6;
 	pSMB->MaxSetupCount = 0;
 	pSMB->Reserved = 0;
 	pSMB->Flags = 0;
 	pSMB->Timeout = 0;
 	pSMB->Reserved2 = 0;
 	pSMB->ParameterOffset =
-	    (int) &(pSMB->InformationLevel) - (int) pSMB->hdr.Protocol;
+	    (int) &(pSMB->Fid) - (int) pSMB->hdr.Protocol;
 	pSMB->DataOffset = pSMB->ParameterOffset + pSMB->ParameterCount;
-	parm_data =
-	    (FILE_END_OF_FILE_INFO *) (((char *) &pSMB->hdr.Protocol) +
-				       pSMB->DataOffset);
+
+	data_offset = (char *) (&pSMB->hdr.Protocol) + pSMB->DataOffset;	
+
+	pSMB->DataCount = sizeof(struct file_end_of_file_info);
+	pSMB->MaxParameterCount = cpu_to_le16(2);
+	pSMB->MaxDataCount = cpu_to_le16(1000);	/* BB find exact max SMB PDU from sess structure BB */
 	pSMB->SetupCount = 1;
 	pSMB->Reserved3 = 0;
-	pSMB->SubCommand = cpu_to_le16(TRANS2_SET_PATH_INFORMATION);
+	pSMB->SubCommand = cpu_to_le16(TRANS2_SET_FILE_INFORMATION);
 	pSMB->ByteCount = 3 /* pad */  + pSMB->ParameterCount + pSMB->DataCount;
-	pSMB->ParameterCount = cpu_to_le16(pSMB->ParameterCount);
 	pSMB->DataCount = cpu_to_le16(pSMB->DataCount);
+	pSMB->ParameterCount = cpu_to_le16(pSMB->ParameterCount);
 	pSMB->TotalDataCount = pSMB->DataCount;
 	pSMB->TotalParameterCount = pSMB->ParameterCount;
-	pSMB->DataOffset = cpu_to_le16(pSMB->DataOffset);
 	pSMB->ParameterOffset = cpu_to_le16(pSMB->ParameterOffset);
-	if (tcon->ses->capabilities & CAP_INFOLEVEL_PASSTHRU)
-		pSMB->InformationLevel =
-		    cpu_to_le16(SMB_SET_FILE_ALLOCATION_INFO2);
-	else
-		pSMB->InformationLevel =
-		    cpu_to_le16(SMB_SET_FILE_ALLOCATION_INFO);
-
+	pSMB->DataOffset = cpu_to_le16(pSMB->DataOffset);
+	parm_data =
+	    (struct file_end_of_file_info *) (((char *) &pSMB->hdr.Protocol) +
+				       pSMB->DataOffset);
+    parm_data->FileSize = size;
+    if(SetAllocation) {
+        if (tcon->ses->capabilities & CAP_INFOLEVEL_PASSTHRU)
+            pSMB->InformationLevel =
+                cpu_to_le16(SMB_SET_FILE_ALLOCATION_INFO2);
+        else
+            pSMB->InformationLevel =
+                cpu_to_le16(SMB_SET_FILE_ALLOCATION_INFO);
+    } else /* Set File Size */  {    
+	    if (tcon->ses->capabilities & CAP_INFOLEVEL_PASSTHRU)
+		    pSMB->InformationLevel =
+		        cpu_to_le16(SMB_SET_FILE_END_OF_FILE_INFO2);
+	    else
+		    pSMB->InformationLevel =
+		        cpu_to_le16(SMB_SET_FILE_END_OF_FILE_INFO);
+    }
 	pSMB->Reserved4 = 0;
 	pSMB->hdr.smb_buf_length += pSMB->ByteCount;
-	parm_data->FileSize = cpu_to_le64(size);
 	pSMB->ByteCount = cpu_to_le16(pSMB->ByteCount);
 	rc = SendReceive(xid, tcon->ses, (struct smb_hdr *) pSMB,
-			 (struct smb_hdr *) pSMBr, &bytes_returned, 2);
+			 (struct smb_hdr *) pSMBr, &bytes_returned, 0);
 	if (rc) {
-		cFYI(1, ("\nSetPathInfo (allocation size) returned %d\n", rc));
+		cFYI(1,
+		     ("\nSend error in SetFileInfo (create symlink) = %d\n",
+		      rc));
 	}
 
 	if (pSMB)
