@@ -115,7 +115,8 @@ cifs_demultiplex_thread(struct TCP_Server_Info *server)
 			return -ENOMEM;
 		}
 		iov.iov_base = smb_buffer;
-		iov.iov_len = sizeof (struct smb_hdr) - 1;	/* 1 byte less since wct is not always returned in error cases */
+		iov.iov_len = sizeof (struct smb_hdr) - 1;	
+        /* 1 byte less above since wct is not always returned in error cases */
 		smb_msg.msg_iov = &iov;
 		smb_msg.msg_iovlen = 1;
 		smb_msg.msg_control = NULL;
@@ -153,7 +154,7 @@ cifs_demultiplex_thread(struct TCP_Server_Info *server)
 			if (temp[0] == (char) 0x85) {
 				iov.iov_base = smb_buffer;
 				iov.iov_len = 4;
-				length = sock_recvmsg(csocket, &smb_msg, 4, 0);	/* BB verify length of keep alive packet BB */
+				length = sock_recvmsg(csocket, &smb_msg, 4, 0);
 				cFYI(0,
 				     ("\nReceived 4 byte keep alive packet "));
 			} else if ((temp[0] == (char) 0x83)
@@ -192,31 +193,21 @@ cifs_demultiplex_thread(struct TCP_Server_Info *server)
 					length = 0;
 					iov.iov_base = smb_buffer;
 					iov.iov_len = pdu_length;
-					for (total_read = 0; total_read < pdu_length; total_read += length) {	/* BB check for buffer overflow with bad pdu_length */
+					for (total_read = 0; total_read < pdu_length; total_read += length) {	
+             /* Should improve check for buffer overflow with bad pdu_length */
 						/*  iov.iov_base = smb_buffer+total_read;
 						   iov.iov_len =  pdu_length-total_read; */
-						length =
-						    sock_recvmsg(csocket,
-								 &smb_msg,
-								 pdu_length
-								 - total_read
-								 /* buflen */
-								 , 0
-								 /* see socket.h */
-						    );
-						/*            cERROR(1,
-						   ("\nFor iovlen %d Length received: %d with total read %d",
-						   iov.iov_len, length,
-						   total_read));       */
+						length = sock_recvmsg(csocket, &smb_msg, 
+                                 pdu_length - total_read, 0);
+         /* cERROR(1,("\nFor iovlen %d Length received: %d with total read %d",
+						   iov.iov_len, length,total_read));       */
 						if (length == 0) {
 							cERROR(1,
 							       ("\nZero length receive when expecting %d ",
-								pdu_length
-								- total_read));
+								pdu_length - total_read));
 							/* BB add reconnect here */
 							break;
-						}
-						/* BB what if error - do we need to check for negative? BB */
+						}						
 					}
 				}
 
@@ -262,11 +253,13 @@ cifs_demultiplex_thread(struct TCP_Server_Info *server)
 		}
 	}
 	/* BB add code to lock SMB sessions while releasing */
-	sock_release(csocket);
-	server->ssocket = NULL;
+	if(server->ssocket) {
+        sock_release(csocket);
+	    server->ssocket = NULL;
+    }
 	set_fs(temp_fs);
-	if (smb_buffer)
-		buf_release(smb_buffer);	/* buffer usually freed in free_mid - but need to free it on error cases or exit case */
+	if (smb_buffer) /* buffer usually freed in free_mid - need to free it on error or exit */
+		buf_release(smb_buffer);
 	if (list_empty(&server->pending_mid_q)) {
 		/* loop through server session structures attached to this and mark them dead */
 		list_for_each(tmp, &GlobalSMBSessionList) {
@@ -279,15 +272,13 @@ cifs_demultiplex_thread(struct TCP_Server_Info *server)
 			}
 		}
 		kfree(server);
-	} else			/* BB need to more gracefully handle the rare negative session response case because response will be still outstanding */
+	} else	/* BB need to more gracefully handle the rare negative session 
+               response case because response will be still outstanding */
 		cERROR(1, ("\nThere are still active MIDs in queue and we are exiting but we can not delete mid_q_entries or TCP_Server_Info structure due to pending requests MEMORY LEAK!!\n "));	/* BB wake up waitors, and/or wait and/or free stale mids and try again? BB */
-/* BB Need to fix bug in error path above - perhaps wait until smb requests time out and then free the tcp per server struct BB */
+/* BB Need to fix bug in error path above - perhaps wait until smb requests
+   time out and then free the tcp per server struct BB */
 
-/* sock_release(csocket); *//* now done in cifs_umount */
 	cFYI(1, ("\nAbout to exit from demultiplex thread\n"));
-
-	/* BB: Do we need to add reconnect logic here? */
-
 	return 0;
 }
 
@@ -348,7 +339,7 @@ parse_mount_options(char *options, char *devname, struct smb_vol *vol)
 				if (strncmp(vol->UNC, "//", 2) == 0) {
 					vol->UNC[0] = '\\';
 					vol->UNC[1] = '\\';
-				} else if (strncmp(vol->UNC, "\\\\", 2) != 0) {	/* target does not begin with // or \\ */
+				} else if (strncmp(vol->UNC, "\\\\", 2) != 0) {	                   
 					printk(KERN_ERR
 					       "CIFS: UNC Path does not begin with // or \\\\");
 					return 1;
@@ -406,7 +397,7 @@ parse_mount_options(char *options, char *devname, struct smb_vol *vol)
 			if (strncmp(vol->UNC, "//", 2) == 0) {
 				vol->UNC[0] = '\\';
 				vol->UNC[1] = '\\';
-			} else if (strncmp(vol->UNC, "\\\\", 2) != 0) {	/* target does not begin with // or \\ */
+			} else if (strncmp(vol->UNC, "\\\\", 2) != 0) {
 				printk(KERN_ERR
 				       "CIFS: UNC Path does not begin with // or \\\\");
 				return 1;
@@ -440,7 +431,7 @@ find_tcp_session(__u32 new_target_ip_addr,
 				if (strncmp
 				    (ses->userName, userName,
 				     MAX_USERNAME_SIZE) == 0)
-					return ses;	/* found an exact match on both SMB session and tcp session */
+					return ses;	/* found exact match on both tcp and SMB sessions */
 			}
 		}
 		/* else tcp and smb sessions need reconnection */
@@ -465,7 +456,7 @@ find_unc(__u32 new_target_ip_addr, char *uncName, char *userName)
 				      s_addr, new_target_ip_addr));
 				if (tcon->ses->server->sockAddr.sin_addr.
 				    s_addr == new_target_ip_addr) {
-					/* BB lock tcon and server and tcp session and increment use count here?? */
+	/* BB lock tcon and server and tcp session and increment use count here? */
 					/* found a match on the TCP session */
 					/* BB check if reconnection needed */
 					cFYI(1,
@@ -545,7 +536,8 @@ ipv4_connect(struct sockaddr_in *psin_server, struct socket **csocket)
 	rc = (*csocket)->ops->connect(*csocket,
 				      (struct sockaddr *) psin_server,
 				      sizeof (struct sockaddr_in), 0
-				      /* BB fix the timeout to be shorter - and find out what are all of the values for these flags? */
+		/* Is there a way to fix a polling timeout -
+         and find out what more of the flags really mean? */
 	    );
 	if (rc < 0) {
 		psin_server->sin_port = htons(RFC1001_PORT);
@@ -605,6 +597,7 @@ cifs_mount(struct super_block *sb, struct cifs_sb_info *cifs_sb,
 {
 	int rc = 0;
 	int xid;
+    int ntlmv2_flag = FALSE;
 	struct socket *csocket;
 	struct sockaddr_in sin_server;
 /*	struct sockaddr_in6 sin_server6; */
@@ -627,7 +620,8 @@ cifs_mount(struct super_block *sb, struct cifs_sb_info *cifs_sb,
 		cFYI(1, ("\nUsername: %s ", volume_info.username));
 
 	} else {
-		cERROR(1, ("\nNo username specified "));	/* BB fix to allow getting user name from alternate locations */
+		cERROR(1, ("\nNo username specified "));	
+        /* Could add ways to allow getting user name from alternate locations */
 	}
 
 	if (volume_info.UNC) {
@@ -663,7 +657,8 @@ cifs_mount(struct super_block *sb, struct cifs_sb_info *cifs_sb,
 			return rc;
 		} else {
 			memset(srvTcp, 0, sizeof (struct TCP_Server_Info));
-			memcpy(&srvTcp->sockAddr, &sin_server, sizeof (struct sockaddr_in));	/* BB fix for ipv6 case */
+			memcpy(&srvTcp->sockAddr, &sin_server, sizeof (struct sockaddr_in));	
+            /* BB Add code for ipv6 case too */
 			srvTcp->ssocket = csocket;
 			init_waitqueue_head(&srvTcp->response_q);
 			INIT_LIST_HEAD(&srvTcp->pending_mid_q);
@@ -725,37 +720,41 @@ cifs_mount(struct super_block *sb, struct cifs_sb_info *cifs_sb,
 				rc = CIFSNTLMSSPNegotiateSessSetup(xid,
 								   pSesInfo,
 								   cryptKey,
-								   volume_info.
-								   domainname,
-								   cifs_sb->
-								   local_nls);
+								   volume_info.domainname,
+                                   &ntlmv2_flag,
+								   cifs_sb->local_nls);
 				if (!rc) {
-					SMBNTencrypt(password_with_pad,
-						     cryptKey,
-						     ntlm_session_key);
-					/* remove the two lines for better security */
-					toUpper(cifs_sb->local_nls,
+                    if(ntlmv2_flag) {
+                        cFYI(1,("\nAble to use the more secure NTLM version 2 password hash"));
+                        /* SMBNTv2encrypt( ...);  */ /* BB fix this up - 
+                        and note that Samba client equivalent looks wrong */
+                    } else
+					    SMBNTencrypt(password_with_pad,cryptKey,ntlm_session_key);
+
+					/* for better security the weaker lanman hash not sent 
+                       in AuthSessSetup so why bother calculating it */
+
+					/* toUpper(cifs_sb->local_nls,
 						password_with_pad);
 					SMBencrypt(password_with_pad,
-						   cryptKey, session_key);
+						   cryptKey, session_key); */
 
 					rc = CIFSNTLMSSPAuthSessSetup(xid,
 								      pSesInfo,
 								      volume_info.
 								      username,
-								      volume_info.
-								      domainname,
+								      volume_info.domainname,
 								      ntlm_session_key,
 								      session_key,
-								      cifs_sb->
-								      local_nls);
+                                      ntlmv2_flag,
+								      cifs_sb->local_nls);
 				}
 			} else {	/* old style NTLM 0.12 session setup */
 				SMBNTencrypt(password_with_pad, cryptKey,
 					     ntlm_session_key);
-				/* we removed following few lines to not send old style password hash ... ever */
-				/* for better security 
-				   toUpper(cifs_sb->local_nls, password_with_pad);
+               /* Removed following few lines to not send old style password 
+                  hash ever - for better security */
+			   /* toUpper(cifs_sb->local_nls, password_with_pad);
 				   SMBencrypt(password_with_pad, cryptKey,session_key); 
 				   dump_mem("\nCIFS (Samba encrypt): ", session_key,CIFS_SESSION_KEY_SIZE); */
 
@@ -825,8 +824,9 @@ cifs_mount(struct super_block *sb, struct cifs_sb_info *cifs_sb,
 
 /* on error free sesinfo and tcon struct if needed */
 	if (rc) {
-		if (tcon)	/* If find_unc succeeded rc == 0 so we can not end here up accidently freeing someone elses tcon struct */
-			tconInfoFree(tcon);	/* BB other cleanup to add? */
+		           /* If find_unc succeeded then rc == 0 so we can not end */
+        if (tcon)  /* up here accidently freeing someone elses tcon struct */
+			tconInfoFree(tcon);
 		if (existingCifsSes == 0) {
 			if (pSesInfo) {
 				if (pSesInfo->server) {
@@ -847,7 +847,7 @@ cifs_mount(struct super_block *sb, struct cifs_sb_info *cifs_sb,
 		cifs_sb->tcon = tcon;
 		tcon->ses = pSesInfo;
 
-		/* we do not care if the following two calls succeed - they are informational only */
+		/* do not care if following two calls succeed - informational only */
 		CIFSSMBQFSDeviceInfo(xid, tcon, cifs_sb->local_nls);
 		CIFSSMBQFSAttributeInfo(xid, tcon, cifs_sb->local_nls);
 		if (tcon->ses->capabilities & CAP_UNIX)
@@ -897,6 +897,9 @@ CIFSSessSetup(unsigned int xid, struct cifsSesInfo *ses, char *user,
 	pSMB->req_no_secext.MaxBufferSize = cpu_to_le16(ses->maxBuf);
 	pSMB->req_no_secext.MaxMpxCount = cpu_to_le16(ses->maxReq);
 
+    if(ses->secMode & (SECMODE_SIGN_REQUIRED | SECMODE_SIGN_ENABLED))
+        smb_buffer->Flags2 |= SMBFLG2_SECURITY_SIGNATURE;
+
 	pSMB->req_no_secext.Capabilities =
 	    CAP_LARGE_FILES | CAP_NT_SMBS | CAP_LEVEL_II_OPLOCKS;
 	if (ses->capabilities & CAP_UNICODE) {
@@ -925,13 +928,16 @@ CIFSSessSetup(unsigned int xid, struct cifsSesInfo *ses, char *user,
 	bcc_ptr += CIFS_SESSION_KEY_SIZE;
 
 	if (ses->capabilities & CAP_UNICODE) {
-		if ((int) bcc_ptr % 2) {	/* must be word aligned for Unicode strings */
+		if ((int) bcc_ptr % 2) {	/* must be word aligned for Unicode */
 			*bcc_ptr = 0;
 			bcc_ptr++;
 		}
-		bytes_returned =
-		    cifs_strtoUCS((wchar_t *) bcc_ptr, user, 100, nls_codepage);
-		bcc_ptr += 2 * bytes_returned;	/* convert num of 16 bit words to bytes */
+        if(user == NULL)
+            bytes_returned = 0; /* skill null user */
+        else
+		    bytes_returned =
+		        cifs_strtoUCS((wchar_t *) bcc_ptr, user, 100, nls_codepage);
+		bcc_ptr += 2 * bytes_returned;	/* convert num 16 bit words to bytes */
 		bcc_ptr += 2;	/* trailing null */
 		if (domain == NULL)
 			bytes_returned =
@@ -958,8 +964,10 @@ CIFSSessSetup(unsigned int xid, struct cifsSesInfo *ses, char *user,
 		bcc_ptr += 2 * bytes_returned;
 		bcc_ptr += 2;
 	} else {
-		strncpy(bcc_ptr, user, 200);
-		bcc_ptr += strnlen(user, 200);
+        if(user != NULL) {                
+		    strncpy(bcc_ptr, user, 200);
+		    bcc_ptr += strnlen(user, 200);
+        }
 		*bcc_ptr = 0;
 		bcc_ptr++;
 		if (domain == NULL) {
@@ -984,21 +992,19 @@ CIFSSessSetup(unsigned int xid, struct cifsSesInfo *ses, char *user,
 
 	rc = SendReceive(xid, ses, smb_buffer, smb_buffer_response,
 			 &bytes_returned, 1);
-	/* dump_mem("\nSessSetup response is: ", smb_buffer_response, 92);  */
+	/* dump_mem("\nSessSetup response is: ", smb_buffer_response, 92);*/
 	if (rc) {
-/*    rc = map_smb_to_linux_error(smb_buffer_response);  *//* done in SendReceive now */
+/* rc = map_smb_to_linux_error(smb_buffer_response); now done in SendReceive */
 	} else if ((smb_buffer_response->WordCount == 3)
 		   || (smb_buffer_response->WordCount == 4)) {
 		pSMBr->resp.Action = le16_to_cpu(pSMBr->resp.Action);
 		if (pSMBr->resp.Action & GUEST_LOGIN)
-			cFYI(1, (" Guest login"));	/* BB do we want to set anything in SesInfo struct ? */
+			cFYI(1, (" Guest login"));	/* do we want to mark SesInfo struct ? */
 		if (ses) {
 			ses->Suid = smb_buffer_response->Uid;	/* UID left in wire format (le) */
 			cFYI(1, ("UID = %d ", ses->Suid));
-
-			/* BB Fix below to make endian neutral !! */
-
-			bcc_ptr = pByteArea(smb_buffer_response);	/* response can have either 3 or 4 word count - Samba sends 3 */
+         /* response can have either 3 or 4 word count - Samba sends 3 */
+			bcc_ptr = pByteArea(smb_buffer_response);	
 			if ((pSMBr->resp.hdr.WordCount == 3)
 			    || ((pSMBr->resp.hdr.WordCount == 4)
 				&& (pSMBr->resp.SecurityBlobLength <
@@ -1021,63 +1027,37 @@ CIFSSessSetup(unsigned int xid, struct cifsSesInfo *ses, char *user,
 					len =
 					    UniStrnlen((wchar_t *) bcc_ptr,
 						       remaining_words - 1);
-/* We look for obvious messed up bcc or strings in response so we do not go off the end since (at least) WIN2K 
-   and Windows XP have a major bug in not null terminating last Unicode string in response  */
-					ses->serverOS =
-					    kcalloc(2 * (len + 1), GFP_KERNEL);
+/* We look for obvious messed up bcc or strings in response so we do not go off
+   the end since (at least) WIN2K and Windows XP have a major bug in not null
+   terminating last Unicode string in response  */
+					ses->serverOS = kcalloc(2 * (len + 1), GFP_KERNEL);
 					cifs_strfromUCS_le(ses->serverOS,
-							   (wchar_t *)
-							   bcc_ptr, len,
-							   nls_codepage);
+							   (wchar_t *)bcc_ptr, len,nls_codepage);
 					bcc_ptr += 2 * (len + 1);
 					remaining_words -= len + 1;
 					ses->serverOS[2 * len] = 0;
 					ses->serverOS[1 + (2 * len)] = 0;
 					if (remaining_words > 0) {
-						len = UniStrnlen((wchar_t *)
-								 bcc_ptr,
+						len = UniStrnlen((wchar_t *)bcc_ptr,
 								 remaining_words
 								 - 1);
-						ses->serverNOS =
-						    kcalloc(2 * (len + 1),
-							    GFP_KERNEL);
-						cifs_strfromUCS_le(ses->
-								   serverNOS,
-								   (wchar_t *)
-								   bcc_ptr,
-								   len,
-								   nls_codepage);
+						ses->serverNOS =kcalloc(2 * (len + 1),GFP_KERNEL);
+						cifs_strfromUCS_le(ses->serverNOS,
+								   (wchar_t *)bcc_ptr,len,nls_codepage);
 						bcc_ptr += 2 * (len + 1);
 						ses->serverNOS[2 * len] = 0;
-						ses->serverNOS[1 +
-							       (2 * len)] = 0;
+						ses->serverNOS[1 + (2 * len)] = 0;
 						remaining_words -= len + 1;
 						if (remaining_words > 0) {
-							len = UniStrnlen((wchar_t *) bcc_ptr, remaining_words);	/* last string is not always null terminated (for e.g. for Windows XP & 2000) */
+							len = UniStrnlen((wchar_t *) bcc_ptr, remaining_words);	
+          /* last string is not always null terminated (for e.g. for Windows XP & 2000) */
 							ses->serverDomain =
-							    kcalloc(2 *
-								    (len +
-								     1),
-								    GFP_KERNEL);
-							cifs_strfromUCS_le
-							    (ses->
-							     serverDomain,
-							     (wchar_t *)
-							     bcc_ptr, len,
-							     nls_codepage);
-							bcc_ptr +=
-							    2 * (len + 1);
-							ses->
-							    serverDomain[2
-									 * len]
-							    = 0;
-							ses->
-							    serverDomain[1
-									 +
-									 (2
-									  *
-									  len)]
-							    = 0;
+							    kcalloc(2*(len+1),GFP_KERNEL);
+							cifs_strfromUCS_le(ses->serverDomain,
+							     (wchar_t *)bcc_ptr,len,nls_codepage);
+							bcc_ptr += 2 * (len + 1);
+							ses->serverDomain[2*len] = 0;
+							ses->serverDomain[1+(2*len)] = 0;
 						} /* else no more room so create dummy domain string */
 						else
 							ses->serverDomain =
@@ -1095,30 +1075,23 @@ CIFSSessSetup(unsigned int xid, struct cifsSesInfo *ses, char *user,
 					if (((int) bcc_ptr + len) - (int)
 					    pByteArea(smb_buffer_response)
 					    <= BCC(smb_buffer_response)) {
-						ses->serverOS =
-						    kcalloc(len + 1,
-							    GFP_KERNEL);
-						strncpy(ses->serverOS,
-							bcc_ptr, len);
+						ses->serverOS = kcalloc(len + 1,GFP_KERNEL);
+						strncpy(ses->serverOS,bcc_ptr, len);
 
 						bcc_ptr += len;
 						bcc_ptr[0] = 0;	/* null terminate the string */
 						bcc_ptr++;
 
 						len = strnlen(bcc_ptr, 1024);
-						ses->serverNOS =
-						    kcalloc(len + 1,
-							    GFP_KERNEL);
-						strncpy(ses->serverNOS, bcc_ptr, len);	/* BB watch for buffer overflow */
+						ses->serverNOS = kcalloc(len + 1,GFP_KERNEL);
+						strncpy(ses->serverNOS, bcc_ptr, len);
 						bcc_ptr += len;
 						bcc_ptr[0] = 0;
 						bcc_ptr++;
 
 						len = strnlen(bcc_ptr, 1024);
-						ses->serverDomain =
-						    kcalloc(len + 1,
-							    GFP_KERNEL);
-						strncpy(ses->serverDomain, bcc_ptr, len);	/* BB watch for buffer overflow */
+						ses->serverDomain = kcalloc(len + 1,GFP_KERNEL);
+						strncpy(ses->serverDomain, bcc_ptr, len);
 						bcc_ptr += len;
 						bcc_ptr[0] = 0;
 						bcc_ptr++;
@@ -1140,7 +1113,7 @@ CIFSSessSetup(unsigned int xid, struct cifsSesInfo *ses, char *user,
 			smb_buffer_response->WordCount));
 		rc = -EIO;
 	}
-	/* BB check wct of 17 and sanity check cryptkeylen before copying */
+	
 	if (smb_buffer)
 		buf_release(smb_buffer);
 	if (smb_buffer_response)
@@ -1187,6 +1160,9 @@ CIFSSpnegoSessSetup(unsigned int xid, struct cifsSesInfo *ses,
 	pSMB->req.MaxBufferSize = cpu_to_le16(ses->maxBuf);
 	pSMB->req.MaxMpxCount = cpu_to_le16(ses->maxReq);
 
+    if(ses->secMode & (SECMODE_SIGN_REQUIRED | SECMODE_SIGN_ENABLED))
+        smb_buffer->Flags2 |= SMBFLG2_SECURITY_SIGNATURE;
+
 	pSMB->req.Capabilities =
 	    CAP_LARGE_FILES | CAP_NT_SMBS | CAP_LEVEL_II_OPLOCKS |
 	    CAP_EXTENDED_SECURITY;
@@ -1312,8 +1288,9 @@ CIFSSpnegoSessSetup(unsigned int xid, struct cifsSesInfo *ses,
 					len =
 					    UniStrnlen((wchar_t *) bcc_ptr,
 						       remaining_words - 1);
-/* We look for obvious messed up bcc or strings in response so we do not go off the end since (at least) WIN2K 
-   and Windows XP have a major bug in not null terminating last Unicode string in response  */
+/* We look for obvious messed up bcc or strings in response so we do not go off
+   the end since (at least) WIN2K and Windows XP have a major bug in not null
+   terminating last Unicode string in response  */
 					ses->serverOS =
 					    kcalloc(2 * (len + 1), GFP_KERNEL);
 					cifs_strfromUCS_le(ses->serverOS,
@@ -1325,60 +1302,38 @@ CIFSSpnegoSessSetup(unsigned int xid, struct cifsSesInfo *ses,
 					ses->serverOS[2 * len] = 0;
 					ses->serverOS[1 + (2 * len)] = 0;
 					if (remaining_words > 0) {
-						len = UniStrnlen((wchar_t *)
-								 bcc_ptr,
+						len = UniStrnlen((wchar_t *)bcc_ptr,
 								 remaining_words
 								 - 1);
 						ses->serverNOS =
 						    kcalloc(2 * (len + 1),
 							    GFP_KERNEL);
-						cifs_strfromUCS_le(ses->
-								   serverNOS,
-								   (wchar_t *)
-								   bcc_ptr,
+						cifs_strfromUCS_le(ses->serverNOS,
+								   (wchar_t *)bcc_ptr,
 								   len,
 								   nls_codepage);
 						bcc_ptr += 2 * (len + 1);
 						ses->serverNOS[2 * len] = 0;
-						ses->serverNOS[1 +
-							       (2 * len)] = 0;
+						ses->serverNOS[1 + (2 * len)] = 0;
 						remaining_words -= len + 1;
 						if (remaining_words > 0) {
-							len = UniStrnlen((wchar_t *) bcc_ptr, remaining_words);	/* last string is not always null terminated (for e.g. for Windows XP & 2000) */
-							ses->serverDomain =
-							    kcalloc(2 *
-								    (len +
-								     1),
-								    GFP_KERNEL);
-							cifs_strfromUCS_le
-							    (ses->
-							     serverDomain,
-							     (wchar_t *)
-							     bcc_ptr, len,
+							len = UniStrnlen((wchar_t *) bcc_ptr, remaining_words);	
+                            /* last string is not always null terminated (for e.g. for Windows XP & 2000) */
+							ses->serverDomain = kcalloc(2*(len+1),GFP_KERNEL);
+							cifs_strfromUCS_le(ses->serverDomain,
+							     (wchar_t *)bcc_ptr, 
+                                 len,
 							     nls_codepage);
-							bcc_ptr +=
-							    2 * (len + 1);
-							ses->
-							    serverDomain[2
-									 * len]
-							    = 0;
-							ses->
-							    serverDomain[1
-									 +
-									 (2
-									  *
-									  len)]
-							    = 0;
+							bcc_ptr += 2*(len+1);
+							ses->serverDomain[2*len] = 0;
+							ses->serverDomain[1+(2*len)] = 0;
 						} /* else no more room so create dummy domain string */
 						else
 							ses->serverDomain =
-							    kcalloc(2,
-								    GFP_KERNEL);
+							    kcalloc(2,GFP_KERNEL);
 					} else {	/* no room so create dummy domain and NOS string */
-						ses->serverDomain =
-						    kcalloc(2, GFP_KERNEL);
-						ses->serverNOS =
-						    kcalloc(2, GFP_KERNEL);
+						ses->serverDomain = kcalloc(2, GFP_KERNEL);
+						ses->serverNOS = kcalloc(2, GFP_KERNEL);
 					}
 				} else {	/* ASCII */
 
@@ -1386,30 +1341,23 @@ CIFSSpnegoSessSetup(unsigned int xid, struct cifsSesInfo *ses,
 					if (((int) bcc_ptr + len) - (int)
 					    pByteArea(smb_buffer_response)
 					    <= BCC(smb_buffer_response)) {
-						ses->serverOS =
-						    kcalloc(len + 1,
-							    GFP_KERNEL);
-						strncpy(ses->serverOS,
-							bcc_ptr, len);
+						ses->serverOS = kcalloc(len + 1, GFP_KERNEL);
+						strncpy(ses->serverOS, bcc_ptr, len);
 
 						bcc_ptr += len;
 						bcc_ptr[0] = 0;	/* null terminate the string */
 						bcc_ptr++;
 
 						len = strnlen(bcc_ptr, 1024);
-						ses->serverNOS =
-						    kcalloc(len + 1,
-							    GFP_KERNEL);
-						strncpy(ses->serverNOS, bcc_ptr, len);	/* BB watch for buffer overflow */
+						ses->serverNOS = kcalloc(len + 1,GFP_KERNEL);
+						strncpy(ses->serverNOS, bcc_ptr, len);
 						bcc_ptr += len;
 						bcc_ptr[0] = 0;
 						bcc_ptr++;
 
 						len = strnlen(bcc_ptr, 1024);
-						ses->serverDomain =
-						    kcalloc(len + 1,
-							    GFP_KERNEL);
-						strncpy(ses->serverDomain, bcc_ptr, len);	/* BB watch for buffer overflow */
+						ses->serverDomain = kcalloc(len + 1, GFP_KERNEL);
+						strncpy(ses->serverDomain, bcc_ptr, len);
 						bcc_ptr += len;
 						bcc_ptr[0] = 0;
 						bcc_ptr++;
@@ -1444,7 +1392,7 @@ int
 CIFSNTLMSSPNegotiateSessSetup(unsigned int xid,
 			      struct cifsSesInfo *ses,
 			      char *challenge_from_server,
-			      char *domain,
+			      char *domain, int * pNTLMv2_flag,
 			      const struct nls_table *nls_codepage)
 {
 	struct smb_hdr *smb_buffer;
@@ -1461,7 +1409,7 @@ CIFSNTLMSSPNegotiateSessSetup(unsigned int xid,
 	PCHALLENGE_MESSAGE SecurityBlob2;
 
 	cFYI(1, ("\nIn NTLMSSP sesssetup (negotiate) "));
-
+    *pNTLMv2_flag = FALSE;
 	smb_buffer = buf_get();
 	if (smb_buffer == 0) {
 		return -ENOMEM;
@@ -1484,6 +1432,9 @@ CIFSNTLMSSPNegotiateSessSetup(unsigned int xid,
 	pSMB->req.MaxBufferSize = cpu_to_le16(ses->maxBuf);
 	pSMB->req.MaxMpxCount = cpu_to_le16(ses->maxReq);
 
+    if(ses->secMode & (SECMODE_SIGN_REQUIRED | SECMODE_SIGN_ENABLED))
+        smb_buffer->Flags2 |= SMBFLG2_SECURITY_SIGNATURE;
+
 	pSMB->req.Capabilities =
 	    CAP_LARGE_FILES | CAP_NT_SMBS | CAP_LEVEL_II_OPLOCKS |
 	    CAP_EXTENDED_SECURITY;
@@ -1505,12 +1456,12 @@ CIFSNTLMSSPNegotiateSessSetup(unsigned int xid,
 	SecurityBlob = (PNEGOTIATE_MESSAGE) bcc_ptr;
 	strncpy(SecurityBlob->Signature, NTLMSSP_SIGNATURE, 8);
 	SecurityBlob->MessageType = NtLmNegotiate;
-	/* SecurityBlob->NegotiateFlags = NTLMSSP_NEGOTIATE_UNICODE | NTLMSSP_REQUEST_TARGET | NTLMSSP_NEGOTIATE_NTLM  | NTLMSSP_NEGOTIATE_LM_KEY | NTLMSSP_NEGOTIATE_KEY_XCH   | NTLMSSP_NEGOTIATE_128 | 0x80000000; */
 	SecurityBlob->NegotiateFlags =
 	    NTLMSSP_NEGOTIATE_UNICODE | NTLMSSP_NEGOTIATE_OEM |
 	    NTLMSSP_REQUEST_TARGET | NTLMSSP_NEGOTIATE_NTLM | 0x80000000 |
-	    NTLMSSP_NEGOTIATE_ALWAYS_SIGN |
-	    /* NTLMSSP_NEGOTIATE_NTLMV2 | */ NTLMSSP_NEGOTIATE_128;
+	    NTLMSSP_NEGOTIATE_ALWAYS_SIGN | NTLMSSP_NEGOTIATE_128;
+    if(ntlmv2_support)
+        SecurityBlob->NegotiateFlags |= NTLMSSP_NEGOTIATE_NTLMV2;
 	/* setup pointers to domain name and workstation name */
 	bcc_ptr += SecurityBlobLength;
 
@@ -1595,9 +1546,11 @@ CIFSNTLMSSPNegotiateSessSetup(unsigned int xid,
 		pSMBr->resp.SecurityBlobLength =
 		    le16_to_cpu(pSMBr->resp.SecurityBlobLength);
 		if (pSMBr->resp.Action & GUEST_LOGIN)
-			cFYI(1, (" Guest login"));	/* BB do we want to set anything in SesInfo struct ? */
+			cFYI(1, (" Guest login"));	
+        /* Do we want to set anything in SesInfo struct when guest login? */
 
-		bcc_ptr = pByteArea(smb_buffer_response);	/* response can have either 3 or 4 word count - Samba sends 3 */
+		bcc_ptr = pByteArea(smb_buffer_response);	
+        /* response can have either 3 or 4 word count - Samba sends 3 */
 
 		SecurityBlob2 = (PCHALLENGE_MESSAGE) bcc_ptr;
 		if (SecurityBlob2->MessageType != NtLmChallenge) {
@@ -1605,11 +1558,8 @@ CIFSNTLMSSPNegotiateSessSetup(unsigned int xid,
 			     ("\nUnexpected NTLMSSP message type received %d",
 			      SecurityBlob2->MessageType));
 		} else if (ses) {
-			ses->Suid = smb_buffer_response->Uid;	/* UID left in wire format (le) */
+			ses->Suid = smb_buffer_response->Uid; /* UID left in le format */ 
 			cFYI(1, ("UID = %d ", ses->Suid));
-
-			/* BB Fix below to make endian neutral !! */
-
 			if ((pSMBr->resp.hdr.WordCount == 3)
 			    || ((pSMBr->resp.hdr.WordCount == 4)
 				&& (pSMBr->resp.SecurityBlobLength <
@@ -1627,7 +1577,8 @@ CIFSNTLMSSPNegotiateSessSetup(unsigned int xid,
 				memcpy(challenge_from_server,
 				       SecurityBlob2->Challenge,
 				       CIFS_CRYPTO_KEY_SIZE);
-
+                if(SecurityBlob2->NegotiateFlags & NTLMSSP_NEGOTIATE_NTLMV2)
+                    *pNTLMv2_flag = TRUE;
 				if (smb_buffer->Flags2 &= SMBFLG2_UNICODE) {
 					if ((int) (bcc_ptr) % 2) {
 						remaining_words =
@@ -1642,8 +1593,9 @@ CIFSNTLMSSPNegotiateSessSetup(unsigned int xid,
 					len =
 					    UniStrnlen((wchar_t *) bcc_ptr,
 						       remaining_words - 1);
-/* We look for obvious messed up bcc or strings in response so we do not go off the end since (at least) WIN2K 
-   and Windows XP have a major bug in not null terminating last Unicode string in response  */
+/* We look for obvious messed up bcc or strings in response so we do not go off
+   the end since (at least) WIN2K and Windows XP have a major bug in not null
+   terminating last Unicode string in response  */
 					ses->serverOS =
 					    kcalloc(2 * (len + 1), GFP_KERNEL);
 					cifs_strfromUCS_le(ses->serverOS,
@@ -1674,7 +1626,8 @@ CIFSNTLMSSPNegotiateSessSetup(unsigned int xid,
 							       (2 * len)] = 0;
 						remaining_words -= len + 1;
 						if (remaining_words > 0) {
-							len = UniStrnlen((wchar_t *) bcc_ptr, remaining_words);	/* last string is not always null terminated (for e.g. for Windows XP & 2000) */
+							len = UniStrnlen((wchar_t *) bcc_ptr, remaining_words);	
+           /* last string is not always null terminated (for e.g. for Windows XP & 2000) */
 							ses->serverDomain =
 							    kcalloc(2 *
 								    (len +
@@ -1723,14 +1676,14 @@ CIFSNTLMSSPNegotiateSessSetup(unsigned int xid,
 							bcc_ptr, len);
 
 						bcc_ptr += len;
-						bcc_ptr[0] = 0;	/* null terminate the string */
+						bcc_ptr[0] = 0;	/* null terminate string */
 						bcc_ptr++;
 
 						len = strnlen(bcc_ptr, 1024);
 						ses->serverNOS =
 						    kcalloc(len + 1,
 							    GFP_KERNEL);
-						strncpy(ses->serverNOS, bcc_ptr, len);	/* BB watch for buffer overflow */
+						strncpy(ses->serverNOS, bcc_ptr, len);
 						bcc_ptr += len;
 						bcc_ptr[0] = 0;
 						bcc_ptr++;
@@ -1739,7 +1692,7 @@ CIFSNTLMSSPNegotiateSessSetup(unsigned int xid,
 						ses->serverDomain =
 						    kcalloc(len + 1,
 							    GFP_KERNEL);
-						strncpy(ses->serverDomain, bcc_ptr, len);	/* BB watch for buffer overflow */
+						strncpy(ses->serverDomain, bcc_ptr, len);	
 						bcc_ptr += len;
 						bcc_ptr[0] = 0;
 						bcc_ptr++;
@@ -1774,7 +1727,7 @@ int
 CIFSNTLMSSPAuthSessSetup(unsigned int xid, struct cifsSesInfo *ses,
 			 char *user, char *domain,
 			 char *ntlm_session_key,
-			 char *lanman_session_key,
+			 char *lanman_session_key, int ntlmv2_flag,
 			 const struct nls_table *nls_codepage)
 {
 	struct smb_hdr *smb_buffer;
@@ -1813,6 +1766,10 @@ CIFSNTLMSSPAuthSessSetup(unsigned int xid, struct cifsSesInfo *ses,
 	pSMB->req.MaxMpxCount = cpu_to_le16(ses->maxReq);
 
 	pSMB->req.hdr.Uid = ses->Suid;
+
+    if(ses->secMode & (SECMODE_SIGN_REQUIRED | SECMODE_SIGN_ENABLED))
+        smb_buffer->Flags2 |= SMBFLG2_SECURITY_SIGNATURE;
+
 	pSMB->req.Capabilities =
 	    CAP_LARGE_FILES | CAP_NT_SMBS | CAP_LEVEL_II_OPLOCKS |
 	    CAP_EXTENDED_SECURITY;
@@ -1838,8 +1795,9 @@ CIFSNTLMSSPAuthSessSetup(unsigned int xid, struct cifsSesInfo *ses,
 	SecurityBlob->NegotiateFlags =
 	    NTLMSSP_NEGOTIATE_UNICODE | NTLMSSP_REQUEST_TARGET |
 	    NTLMSSP_NEGOTIATE_NTLM | NTLMSSP_NEGOTIATE_TARGET_INFO |
-	    0x80000000 | NTLMSSP_NEGOTIATE_ALWAYS_SIGN
-	    /* | NTLMSSP_NEGOTIATE_NTLMV2 */  | NTLMSSP_NEGOTIATE_128;
+	    0x80000000 | NTLMSSP_NEGOTIATE_ALWAYS_SIGN | NTLMSSP_NEGOTIATE_128;
+    if(ntlmv2_flag)
+        SecurityBlob->NegotiateFlags |= NTLMSSP_NEGOTIATE_NTLMV2;
 
 /* setup pointers to domain name and workstation name */
 
@@ -2004,11 +1962,9 @@ CIFSNTLMSSPAuthSessSetup(unsigned int xid, struct cifsSesInfo *ses,
 			cFYI(1,
 			     ("Does UID on challenge %d match auth response UID %d ",
 			      ses->Suid, smb_buffer_response->Uid));
-			ses->Suid = smb_buffer_response->Uid;	/* UID left in wire format (le) */
-			bcc_ptr = pByteArea(smb_buffer_response);	/* response can have either 3 or 4 word count - Samba sends 3 */
-
-			/* BB Fix below to make endian neutral !! */
-
+			ses->Suid = smb_buffer_response->Uid; /* UID left in wire format */
+			bcc_ptr = pByteArea(smb_buffer_response);	
+            /* response can have either 3 or 4 word count - Samba sends 3 */
 			if ((pSMBr->resp.hdr.WordCount == 3)
 			    || ((pSMBr->resp.hdr.WordCount == 4)
 				&& (pSMBr->resp.SecurityBlobLength <
@@ -2031,15 +1987,13 @@ CIFSNTLMSSPAuthSessSetup(unsigned int xid, struct cifsSesInfo *ses,
 						     - 1) / 2;
 						bcc_ptr++;	/* Unicode strings must be word aligned */
 					} else {
-						remaining_words =
-						    BCC
-						    (smb_buffer_response) / 2;
+						remaining_words = BCC(smb_buffer_response) / 2;
 					}
 					len =
-					    UniStrnlen((wchar_t *) bcc_ptr,
-						       remaining_words - 1);
-/* We look for obvious messed up bcc or strings in response so we do not go off the end since (at least) WIN2K 
-   and Windows XP have a major bug in not null terminating last Unicode string in response  */
+					    UniStrnlen((wchar_t *) bcc_ptr,remaining_words - 1);
+/* We look for obvious messed up bcc or strings in response so we do not go off
+  the end since (at least) WIN2K and Windows XP have a major bug in not null
+  terminating last Unicode string in response  */
 					ses->serverOS =
 					    kcalloc(2 * (len + 1), GFP_KERNEL);
 					cifs_strfromUCS_le(ses->serverOS,
@@ -2066,11 +2020,11 @@ CIFSNTLMSSPAuthSessSetup(unsigned int xid, struct cifsSesInfo *ses,
 								   nls_codepage);
 						bcc_ptr += 2 * (len + 1);
 						ses->serverNOS[2 * len] = 0;
-						ses->serverNOS[1 +
-							       (2 * len)] = 0;
+						ses->serverNOS[1+(2*len)] = 0;
 						remaining_words -= len + 1;
 						if (remaining_words > 0) {
-							len = UniStrnlen((wchar_t *) bcc_ptr, remaining_words);	/* last string is not always null terminated (for e.g. for Windows XP & 2000) */
+							len = UniStrnlen((wchar_t *) bcc_ptr, remaining_words);	
+     /* last string is not always null terminated (for e.g. for Windows XP & 2000) */
 							ses->serverDomain =
 							    kcalloc(2 *
 								    (len +
@@ -2097,45 +2051,34 @@ CIFSNTLMSSPAuthSessSetup(unsigned int xid, struct cifsSesInfo *ses,
 							    = 0;
 						} /* else no more room so create dummy domain string */
 						else
-							ses->serverDomain =
-							    kcalloc(2,
-								    GFP_KERNEL);
+							ses->serverDomain = kcalloc(2,GFP_KERNEL);
 					} else {	/* no room so create dummy domain and NOS string */
-						ses->serverDomain =
-						    kcalloc(2, GFP_KERNEL);
-						ses->serverNOS =
-						    kcalloc(2, GFP_KERNEL);
+						ses->serverDomain = kcalloc(2, GFP_KERNEL);
+						ses->serverNOS = kcalloc(2, GFP_KERNEL);
 					}
 				} else {	/* ASCII */
 
 					len = strnlen(bcc_ptr, 1024);
-					if (((int) bcc_ptr + len) - (int)
-					    pByteArea(smb_buffer_response)
-					    <= BCC(smb_buffer_response)) {
-						ses->serverOS =
-						    kcalloc(len + 1,
-							    GFP_KERNEL);
-						strncpy(ses->serverOS,
-							bcc_ptr, len);
+					if (((int) bcc_ptr + len) - 
+                        (int) pByteArea(smb_buffer_response) 
+                            <= BCC(smb_buffer_response)) {
+						ses->serverOS = kcalloc(len + 1,GFP_KERNEL);
+						strncpy(ses->serverOS,bcc_ptr, len);
 
 						bcc_ptr += len;
 						bcc_ptr[0] = 0;	/* null terminate the string */
 						bcc_ptr++;
 
 						len = strnlen(bcc_ptr, 1024);
-						ses->serverNOS =
-						    kcalloc(len + 1,
-							    GFP_KERNEL);
-						strncpy(ses->serverNOS, bcc_ptr, len);	/* BB watch for buffer overflow */
+						ses->serverNOS = kcalloc(len+1,GFP_KERNEL);
+						strncpy(ses->serverNOS, bcc_ptr, len);	
 						bcc_ptr += len;
 						bcc_ptr[0] = 0;
 						bcc_ptr++;
 
 						len = strnlen(bcc_ptr, 1024);
-						ses->serverDomain =
-						    kcalloc(len + 1,
-							    GFP_KERNEL);
-						strncpy(ses->serverDomain, bcc_ptr, len);	/* BB watch for buffer overflow */
+						ses->serverDomain = kcalloc(len+1,GFP_KERNEL);
+						strncpy(ses->serverDomain, bcc_ptr, len);
 						bcc_ptr += len;
 						bcc_ptr[0] = 0;
 						bcc_ptr++;
@@ -2204,6 +2147,9 @@ CIFSTCon(unsigned int xid, struct cifsSesInfo *ses,
 	bcc_ptr = &(pSMB->Password[0]);
 	bcc_ptr++;		/* skip password */
 
+    if(ses->secMode & (SECMODE_SIGN_REQUIRED | SECMODE_SIGN_ENABLED))
+        smb_buffer->Flags2 |= SMBFLG2_SECURITY_SIGNATURE;
+
 	if (ses->capabilities & CAP_STATUS32) {
 		smb_buffer->Flags2 |= SMBFLG2_ERR_STATUS;
 	}
@@ -2230,13 +2176,14 @@ CIFSTCon(unsigned int xid, struct cifsSesInfo *ses,
 
 	rc = SendReceive(xid, ses, smb_buffer, smb_buffer_response, &length, 0);
 
-	/* if (rc) rc = map_smb_to_linux_error(smb_buffer_response); not needed now done in SendReceive */
+	/* if (rc) rc = map_smb_to_linux_error(smb_buffer_response); now done in SendReceive */
 	if ((rc == 0) && (tcon != NULL)) {
 		tcon->tid = smb_buffer_response->Tid;
 		bcc_ptr = pByteArea(smb_buffer_response);
 		length = strnlen(bcc_ptr, BCC(smb_buffer_response) - 2);
-		bcc_ptr += length + 1;	/* skip service field (NB: this field is always ASCII) */
-		strncpy(tcon->treeName, tree, MAX_TREE_SIZE);
+        /* skip service field (NB: this field is always ASCII) */
+		bcc_ptr += length + 1;	
+        strncpy(tcon->treeName, tree, MAX_TREE_SIZE);
 		if (smb_buffer->Flags2 &= SMBFLG2_UNICODE) {
 			length = UniStrnlen((wchar_t *) bcc_ptr, 512);
 			if (((int) bcc_ptr + (2 * length)) -
@@ -2268,7 +2215,8 @@ CIFSTCon(unsigned int xid, struct cifsSesInfo *ses,
 		tcon->Flags = le16_to_cpu(pSMBr->OptionalSupport);
 		cFYI(1, ("\nTcon flags: 0x%x ", tcon->Flags));
 	} else if ((rc == 0) && tcon == NULL) {
-		ses->ipc_tid = smb_buffer_response->Tid;	/* all we need to save for IPC$ connection */
+        /* all we need to save for IPC$ connection */
+		ses->ipc_tid = smb_buffer_response->Tid;
 	}
 
 	if (smb_buffer)
@@ -2287,7 +2235,7 @@ cifs_umount(struct super_block *sb, struct cifs_sb_info *cifs_sb)
 
 	xid = GetXid();
 	if (cifs_sb->tcon) {
-		ses = cifs_sb->tcon->ses;	/* have to save ptr to ses struct before we delete tcon! */
+		ses = cifs_sb->tcon->ses; /* have to save ptr to ses struct before delete tcon!*/
 		rc = CIFSSMBTDis(xid, cifs_sb->tcon);
 		if (rc == -EBUSY) {
 			FreeXid(xid);
@@ -2302,8 +2250,15 @@ cifs_umount(struct super_block *sb, struct cifs_sb_info *cifs_sb)
 				FreeXid(xid);
 				return 0;
 			}
-			wake_up_process(ses->server->tsk);
+    	 /* wake_up_process(ses->server->tsk);*/ /* was worth a try */
 			schedule_timeout(HZ / 4);	/* give captive thread time to exit */
+            if((ses->server) && (ses->server->ssocket)) {            
+                cFYI(1,("\nWaking up socket by closing it "));
+                /* No luck figuring out a better way to_close socket */
+                ses->server->ssocket->sk->prot->close(ses->server->ssocket->sk,0);
+                ses->server->ssocket = NULL;   /* serialize better */
+                /* sock_wake_async(ses->server->ssocket,3,POLL_HUP); */
+            }
 		} else
 			cFYI(1, ("\nNo session or bad tcon"));
 	}
@@ -2311,12 +2266,15 @@ cifs_umount(struct super_block *sb, struct cifs_sb_info *cifs_sb)
 	cifs_sb->tcon = NULL;
 	if (ses) {
 		schedule_timeout(HZ / 2);
-		/*    if (ses->server)
-   kfree(ses->server); *//* freed in demultiplex thread */
+		/* if ((ses->server) && (ses->server->ssocket)) {
+               cFYI(1,("\nReleasing socket "));        
+               sock_release(ses->server->ssocket); 
+               kfree(ses->server); 
+          } */ 
 	}
 	if (ses)
 		sesInfoFree(ses);
 
 	FreeXid(xid);
 	return rc;		/* BB check if we should always return zero here */
-}
+} 
