@@ -42,6 +42,8 @@ cifs_open(struct inode *inode, struct file *file)
 	int xid, oplock;
 	struct cifs_sb_info *cifs_sb;
 	struct cifsTconInfo *pTcon;
+    struct cifsFileInfo *pCifsFile;
+    struct cifsInodeInfo *pCifsInode;
 	char *full_path = NULL;
 	int desiredAccess = 0x20197;
 	int disposition = FILE_OPEN;
@@ -62,26 +64,9 @@ cifs_open(struct inode *inode, struct file *file)
 	else if ((file->f_flags & O_ACCMODE) == O_RDWR)
 		desiredAccess = GENERIC_ALL;
 
-/*	if ((file->f_flags & O_ACCMODE) == O_RDONLY)   
-        desiredAccess =
-		    FILE_READ_DATA | FILE_READ_ATTRIBUTES | FILE_EXECUTE |
-		    READ_CONTROL;
-	else if ((file->f_flags & O_ACCMODE) == O_WRONLY)
-		desiredAccess =
-		    FILE_WRITE_DATA | FILE_WRITE_ATTRIBUTES |
-		    FILE_APPEND_DATA;
-	else if ((file->f_flags & O_ACCMODE) == O_RDWR)
-		desiredAccess =
-		    FILE_WRITE_DATA | FILE_READ_DATA |
-		    FILE_WRITE_ATTRIBUTES | FILE_READ_ATTRIBUTES |
-		    FILE_APPEND_DATA | FILE_EXECUTE | READ_CONTROL;  */
+/* BB check other flags carefully to find equivalent NTCreateX flags */
 
-/* BB go through these flags carefully to find equivalent NTCreateX flags */
 /*
-#define O_ACCMODE	   0003
-#define O_RDONLY	     00
-#define O_WRONLY	     01
-#define O_RDWR		     02
 #define O_CREAT		   0100	
 #define O_EXCL		   0200	
 #define O_NOCTTY	   0400	
@@ -98,7 +83,7 @@ cifs_open(struct inode *inode, struct file *file)
 #define O_ATOMICLOOKUP	01000000 */
 
 	if (file->f_flags & O_CREAT)
-		disposition = FILE_OPEN_IF;	/* open if exists, create if does not exist */
+		disposition = FILE_OVERWRITE;
 
 	/* BB finish adding in oplock support BB */
 	if (oplockEnabled)
@@ -107,10 +92,9 @@ cifs_open(struct inode *inode, struct file *file)
 		oplock = FALSE;
 
 	/* BB pass O_SYNC flag through on file attributes .. BB */
-	rc = CIFSSMBOpen(xid, pTcon, full_path, disposition, desiredAccess,
-			 0x3FF
-			 /* i.e. 777 dummy value, ignored for time being */
-			 , &netfid, &oplock, cifs_sb->local_nls);
+	rc = CIFSSMBOpen(xid, pTcon, full_path, disposition, desiredAccess,			 
+			 -1 /* i.e. dummy value, ignored for time being */,
+			 &netfid, &oplock, cifs_sb->local_nls);
 	if (rc) {
 		cFYI(1, ("\ncifs_open returned 0x%x ", rc));
 		cFYI(1, (" oplock: %d ", oplock));
@@ -120,8 +104,28 @@ cifs_open(struct inode *inode, struct file *file)
 		if (file->private_data) {
 			memset(file->private_data, 0,
 			       sizeof (struct cifsFileInfo));
-			((struct cifsFileInfo *) file->private_data)->
-			    netfid = netfid;
+            pCifsFile = (struct cifsFileInfo *) file->private_data;
+			pCifsFile->netfid = netfid;
+			pCifsFile->pid = current->pid;
+            list_add(&pCifsFile->tlist,&pTcon->openFileList);
+            pCifsInode = CIFS_I(file->f_dentry->d_inode);
+            list_add(&pCifsFile->flist,&pCifsInode->openFileList);
+            if(file->f_flags & O_CREAT) {           
+                 /* time to set mode which we can not
+                 set earlier due to problems creating new read-only files */
+                if (cifs_sb->tcon->ses->capabilities & CAP_UNIX)                
+                    CIFSSMBUnixSetPerms(xid, pTcon, full_path, inode->i_mode,
+                                        0xFFFFFFFFFFFFFFFF,  
+                                        0xFFFFFFFFFFFFFFFF,
+                                        cifs_sb->local_nls);
+                else {/* BB to be implemented via Windows secrty descriptors*/
+         /* eg CIFSSMBWinSetPerms(xid,pTcon,full_path,mode,-1,-1,local_nls);*/
+                    /* in the meantime we could set the r/o dos attribute 
+                       when perms are e.g.:   
+                            mode & 0222 == 0 */
+                }
+            }
+
 		}
 	}
 
@@ -147,6 +151,8 @@ cifs_close(struct inode *inode, struct file *file)
 
 	cifs_sb = (struct cifs_sb_info *) (&(inode->i_sb->u));
 	pTcon = cifs_sb->tcon;
+    list_del(&pSMBFile->flist);
+    list_del(&pSMBFile->tlist);
 
 	if (pSMBFile) {
 		rc = CIFSSMBClose(xid, pTcon, pSMBFile->netfid);
@@ -267,7 +273,7 @@ cifs_lock(struct file *file, int cmd, struct file_lock *pfLock)
 
 		} else {
 			/* if rc == ERR_SHARING_VIOLATION ? */
-			rc = 0;	/* do not change lock type to unlock since the range is in use */
+			rc = 0;	/* do not change lock type to unlock since range in use */
 		}
 
 		FreeXid(xid);
@@ -331,9 +337,9 @@ cifs_write(struct file * file, const char *write_data,
 			}
 		} else
 			*poffset += bytes_written;
-		long_op = FALSE;	/* subsequent writes should be fast - 15 seconds is plenty of time */
+		long_op = FALSE; /* subsequent writes fast - 15 seconds is plenty */
 	}
-	file->f_dentry->d_inode->i_ctime = file->f_dentry->d_inode->i_mtime = CURRENT_TIME;	/* BB is this right? */
+	file->f_dentry->d_inode->i_ctime = file->f_dentry->d_inode->i_mtime = CURRENT_TIME;
 	if (bytes_written > 0) {
 		if (*poffset > file->f_dentry->d_inode->i_size)
 			file->f_dentry->d_inode->i_size = *poffset;
@@ -387,7 +393,7 @@ static int
 cifs_prepare_write(struct file *file, struct page *page, unsigned offset,
 		   unsigned to)
 {
-	return 0;		/* eventually add code to flush incompatible requests if any */
+	return 0;	/* eventually add code to flush any incompatible requests */
 }
 
 static int
@@ -397,7 +403,10 @@ cifs_commit_write(struct file *file, struct page *page, unsigned offset,
 	long rc = -EFAULT;
 
 	lock_kernel();
-/*	status = cifs_updatepage(file, page, offset, to-offset); *//* BB add - do we really need to lock the kernel here for so long ? */
+/*	status = cifs_updatepage(file, page, offset, to-offset); */
+
+/* BB add - do we really need to lock the kernel here for so long ? */
+
 	unlock_kernel();
 	return rc;
 }
@@ -558,9 +567,9 @@ fill_in_inode(struct inode *tmp_inode,
 	} else {
 		*pobject_type = DT_REG;
 		tmp_inode->i_mode |= S_IFREG;
-	}			/* BB add code here - to validate if device or weird share type or odd device type? BB */
+	}/* could add code here - to validate if device or weird share type? */
 
-	/* can not fill in nlink here as we can in both qpathinfo versions and in unix search */
+	/* can not fill in nlink here as in qpathinfo version and in Unx search */
 	tmp_inode->i_size = pfindData->EndOfFile;
 	tmp_inode->i_blocks =
 	    do_div(pfindData->AllocationSize, tmp_inode->i_blksize);
@@ -568,7 +577,7 @@ fill_in_inode(struct inode *tmp_inode,
 	     ("\nFinddata alloc size (from smb) %lld",
 	      pfindData->AllocationSize));
 	if (pfindData->AllocationSize < pfindData->EndOfFile)
-		cFYI(1, ("\nServer inconsistency Error: it says allocation size less than end of file "));	/* BB remove */
+		cFYI(1, ("\nServer inconsistency Error: it says allocation size less than end of file "));
 	cFYI(1,
 	     ("\nCIFS FFIRST: Size %ld and blocks %ld and blocksize %ld",
 	      (unsigned long) tmp_inode->i_size, tmp_inode->i_blocks,
@@ -591,7 +600,6 @@ fill_in_inode(struct inode *tmp_inode,
 	}
 }
 
-/* BB make endian neutral */
 void
 unix_fill_in_inode(struct inode *tmp_inode,
 		   FILE_UNIX_INFO * pfindData, int *pobject_type)
