@@ -344,7 +344,7 @@ cifs_mkdir(struct inode *inode, struct dentry *direntry, int mode)
 	char *full_path = NULL;
 	struct inode *newinode = NULL;
 
-	cFYI(1, ("In cifs_mkdir, inode = 0x%p\n", inode));
+	cFYI(1, ("In cifs_mkdir, mode = 0x%x inode = 0x%p\n", mode, inode));
 
 	xid = GetXid();
 
@@ -368,6 +368,15 @@ cifs_mkdir(struct inode *inode, struct dentry *direntry, int mode)
 		direntry->d_op = &cifs_dentry_ops;
 		d_instantiate(direntry, newinode);
 		direntry->d_inode->i_nlink = 2;
+        if (cifs_sb->tcon->ses->capabilities & CAP_UNIX)                
+            CIFSSMBUnixSetPerms(xid, pTcon, full_path, mode,
+                        0xFFFFFFFFFFFFFFFF,  
+                        0xFFFFFFFFFFFFFFFF,
+                        cifs_sb->local_nls);
+        else { /* BB to be implemented via Windows secrty descriptors*/
+        /* eg CIFSSMBWinSetPerms(xid,pTcon,full_path,mode,-1,-1,local_nls);*/
+        }
+
 	}
 	if (full_path)
 		kfree(full_path);
@@ -500,10 +509,13 @@ cifs_truncate_file(struct inode *inode)
 {				/* BB remove - may not need this function after all BB */
 	int xid;
 	int rc = 0;
+    struct list_head *tmp;
+    struct cifsFileInfo *open_file = NULL;
 	struct cifs_sb_info *cifs_sb;
 	struct cifsTconInfo *pTcon;
+	struct cifsInodeInfo *cifsInode;
 	struct dentry *dirent;
-	char *full_path = NULL;
+	char *full_path = NULL;   
 
 	xid = GetXid();
 
@@ -520,14 +532,22 @@ cifs_truncate_file(struct inode *inode)
 	dirent = list_entry(inode->i_dentry.next, struct dentry, d_alias);
 	if (dirent) {
 		full_path = build_path_from_dentry(dirent);
-		rc = CIFSSMBSetEOF(xid, pTcon, full_path, inode->i_size,
+		rc = CIFSSMBSetEOF(xid, pTcon, full_path, inode->i_size,FALSE,
 				   cifs_sb->local_nls);
+        if(rc == -ETXTBSY) {        
+            cifsInode = CIFS_I(inode);
+	        list_for_each(tmp,&cifsInode->openFileList) {
+		        open_file = list_entry(tmp, struct cifsFileInfo, flist);           
+                /* We could check if file is open for writing first and 
+                   also we could also override the smb pid with the pid 
+                   of the file opener when sending the CIFS request */
+                rc = CIFSSMBSetFileSize(xid, pTcon, inode->i_size,
+                                        open_file->netfid,open_file->pid,FALSE);
+            }
+        }
 		if (!rc)
-			rc = CIFSSMBSetAllocationSize(xid, pTcon,
-						      full_path,
-						      inode->i_size,
-						      cifs_sb->local_nls);
-
+			CIFSSMBSetEOF(xid,pTcon,full_path,inode->i_size,TRUE,cifs_sb->local_nls);
+           /* allocation size setting seems optional so we ignore return code */
 	}
 	if (full_path)
 		kfree(full_path);
@@ -543,6 +563,8 @@ cifs_setattr(struct dentry *direntry, struct iattr *attrs)
 	struct cifsTconInfo *pTcon;
 	char *full_path = NULL;
 	int rc = -EACCES;
+    struct list_head *tmp;
+    struct cifsFileInfo *open_file = NULL;
 	FILE_BASIC_INFO time_buf;
 	int set_time = FALSE;
 	__u64 mode = 0xFFFFFFFFFFFFFFFF;
@@ -559,15 +581,26 @@ cifs_setattr(struct dentry *direntry, struct iattr *attrs)
 	pTcon = cifs_sb->tcon;
 
 	full_path = build_path_from_dentry(direntry);
+	cifsInode = CIFS_I(direntry->d_inode);
 
-	/* BB refresh inode from server now ? BB */
+	/* BB check if we need to refresh inode from server now ? BB */
 
 	cFYI(1, ("\nChanging attributes 0x%x", attrs->ia_valid));
 
 	if (attrs->ia_valid & ATTR_SIZE) {
-		rc = CIFSSMBSetEOF(xid, pTcon, full_path, attrs->ia_size,
+		rc = CIFSSMBSetEOF(xid, pTcon, full_path, attrs->ia_size,FALSE,
 				   cifs_sb->local_nls);
-		CIFSSMBSetAllocationSize(xid, pTcon, full_path, attrs->ia_size, cifs_sb->local_nls);	/* dont really care whether works */
+        if(rc == -ETXTBSY) {
+            list_for_each(tmp,&(cifsInode->openFileList)) {
+                open_file = list_entry(tmp, struct cifsFileInfo, flist);           
+    /* We could check if file is open for writing first */
+                rc = CIFSSMBSetFileSize(xid, pTcon, attrs->ia_size,
+                                        open_file->netfid,open_file->pid,FALSE);
+            }
+        }
+        /*  Set Allocation Size of file - might not even need to call the
+            following but might as well and it does not hurt if it fails */
+		CIFSSMBSetEOF(xid, pTcon, full_path, attrs->ia_size, TRUE, cifs_sb->local_nls);
 		if (rc == 0)
 			vmtruncate(direntry->d_inode, attrs->ia_size);
 		/* BB add special case to handle sharing violation (due to Samba bug)
@@ -625,7 +658,6 @@ cifs_setattr(struct dentry *direntry, struct iattr *attrs)
 		rc = CIFSSMBSetTimes(xid, pTcon, full_path, &time_buf,
 				     cifs_sb->local_nls);
 	}
-	cifsInode = CIFS_I(direntry->d_inode);
 	cifsInode->time = 0;	/* will force revalidate to go get attributes when needed */
 
 	if (full_path)
